#include <time.h>
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "incs.h"
#include "getArgs.h"
#include "tst_data.h"
#include "utils.h"
#include "class_tree.h"


extern "C" int gethostname(char*, ...);

#define FALSE 0
#define TRUE 1
#define DEF_NINP 1
#define DEF_N_OUTPUT 2
#define DEF_N_EXPERTS 2
#define DEF_CV_EPOCH 10
#define DEF_NTEST 100
#define DEF_DEBUG 0
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 1
#endif

int main(int argc, char **argv) 
{
  char* progName = argv[0];  
  char* version = "0.94";
  getArgs comline = getArgs( &argc, argv );
 // process the command line
  bool more_help = (bool) FALSE;
  if(argc == 1){
    more_help = (bool) TRUE;
  }

  bool help = comline.switchArg( "-help",
				 "\tDisplay command line usage",
				 false );

  int debug = comline.intArg( "-debug",
			      "\tdebug level\t\t",
			      DEF_DEBUG );
  int ntest = comline.intArg( "-n_test",
			      "number of test points",
			      DEF_NTEST );
  char* in_file = comline.stringArg( "-in_file",
				      "Weights file \t",
				      "-");
  
  char* test_file = comline.stringArg( "-test_file",
					"Test data file\t",
					"-");
  char* res_file = comline.stringArg( "-res_file", 
				      "Results file (Matlab format)",
				      "hme_res.m");

  // show command line usage
  if ( help || more_help) {
    comline.printUsage( "test [options]" );
    exit( EXIT_SUCCESS );
  }

  comline.finalCheck( 0, "test [options]" );
  if(strcmp(test_file, "-")==0)
    {
      error(1, "You must specify a test data file");
    }
  if(strcmp(in_file, "-")==0)
    {
      error(1, "You must specify an input weights file");
    }
  char hostname[1024];
  time_t now;
  time(&now);
  assert(gethostname(hostname, 1024)==0);
  printf("Program %s version %s started on %s \n at %s \n",
		 progName,version, hostname,ctime(&now));
  int ncomm;
	char* in_file_comm = CharVector(1024);
	char* extn = CharVector(1024);
	int n_out = 0;
	real** overall_outputs = NULL;
	real** inputs;
	real** outputs = NULL;
	real** mom2 = NULL;
	char* spextn = CharVector(1024);
	strcpy(spextn,".spec");
	char* spec_fn = CharVector(1024);
	strcpy(spec_fn, in_file);
	strcat(spec_fn, spextn);
	FILE* spec_fp = openFile(spec_fn, "r");
 	fscanf(spec_fp, "%d\n", &ncomm);
	fclose(spec_fp);
	printf("ncomm = %d\n", ncomm);
	for(int comm = 0; comm < ncomm; comm ++) {
	  FILE* data_fp;
	  data_fp = openFile(test_file, "r");
	  sprintf(extn, ".%d", comm);
	  strcpy(in_file_comm, in_file);
	  strcat(in_file_comm, extn);
	  printf("loading model from %s\n", in_file_comm);
	  FILE* in_fp = openFile(in_file_comm, "r");
	  Class_Tree* class_tree;
	  class_tree = new Class_Tree(ntest, in_fp, debug);
	  
	  fclose(in_fp);
	  	  
	  int n_input = class_tree->num_input()-1;
	  int n_output = class_tree->num_output();
	  n_out = n_output;
	  TST_Data* test_data;
	  test_data = new TST_Data(data_fp, ntest, n_input);
	  inputs = test_data->get_inputs(0);
	  if(outputs == NULL) 
		outputs = RealMatrix2(ntest, n_output);
	  if(mom2 == NULL)
		mom2 = RealMatrix2(ntest, n_output);
	  if(overall_outputs == NULL) {
		overall_outputs = RealMatrix2(ntest, n_output);
		printf("allocating new outputs\n");
		for(int point = 0; point < ntest; point ++){
		  for(int op = 0; op < n_output; op ++){
			overall_outputs[point][op] = 0.0;
		  }
		}
	  }
	  class_tree->go_forward_batch(ntest, inputs, outputs, mom2);
	  fclose(data_fp);
	  delete test_data;
	  test_data = (TST_Data*) NULL;
	  for(int n = 0; n < ntest; n ++) {
		for(int o = 0; o < n_out; o ++) {
		  overall_outputs[n][o] += outputs[n][o];
		}
	  }
 	  delete class_tree;
	  class_tree = (Class_Tree*) NULL;
	}
	for(int n1 = 0; n1 < ntest; n1 ++) {
	  for(int o = 0; o < n_out; o ++) {
		overall_outputs[n1][o] /= ncomm;
	  }
	}
	FILE* fp_res =  openFile(res_file, "w");
	for(int n = 0; n < ntest; n ++) {
	  int best = 0;
	  real best_val = -FLT_MAX;
	  for(int o = 0; o < n_out; o ++) {
		if(overall_outputs[n][o] > best_val) {
		  best_val = overall_outputs[n][o];
		  best = o;
		}
	  }
	  for(int o = 0; o < n_out; o ++) {
		if(o == best)
		  fprintf(fp_res, "%d ", 1);
		else
		  fprintf(fp_res, "%d ", 0);
	  }
	  fprintf(fp_res, "\n");
	}
	fclose(fp_res);  

  time(&now);
  printf("Program %s ended at %s \n",progName,ctime(&now));
  
}







