#include <time.h>
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "incs.h"
#include "getArgs.h"
#include "data.h"
#include "utils.h"
#include "reg_tree.h"
#include "net_utils.h"
#include "functions.h"

extern "C" int gethostname(char*, ...);

#define DEF_MSTEPS 10
#define DEF_TOL 0.001
#define DEF_ITMAX 10
#define DEF_PRIOR_VAR 1.0
#define DEF_LAMBDA 2
#define FALSE 0
#define TRUE 1
#define DEF_NINP 1
#define DEF_N_OUTPUT 2
#define DEF_DEPTH 1
#define DEF_SEED -1
#define DEF_MAX_EPOCHS 200
#define DEF_GATE_RAND 1.0
#define DEF_XPRT_RAND 1.0
#define DEF_NTRAIN 100
#define DEF_NCV 00
#define DEF_DEBUG 0
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 1
#endif
int main(int argc, char **argv) 
{
  char* progName = argv[0];  
  char* version = "0.93";
  getArgs comline = getArgs( &argc, argv );
  // process the command line
  bool more_help = (bool) FALSE;
  if(argc == 1){
    more_help = (bool) TRUE;
  }
  bool help = comline.switchArg( "-help",
				 "\tDisplay command line usage",
				 false );
  int seed = comline.intArg( "-seed",
			     "\trandom seed\t\t",
			     DEF_SEED );
  if(seed < -1)
    error(1, "seed = %d : seed must be greater than 0", seed);

  int n_input = comline.intArg( "-n_input",
				"input vector dimension\t",
				DEF_NINP );
  if(n_input < 1)
    error(1, "n_input = %d : Number of inputs must be > 0", n_input);
  
  int n_output = comline.intArg( "-n_output",
				 "target vector dimension\t",
				 DEF_N_OUTPUT );

  int npoints = comline.intArg( "-n_train",
			       "number of training points",
			       DEF_NTRAIN );
  char* out_file = comline.stringArg( "-out_file",
									  "Weights out file stem\t",
									  "-");
  char* train_file = comline.stringArg( "-train_file",
					"Training data file\t",
					"-");
  // show command line usage
  if ( help || more_help) {
    comline.printUsage( "test [options]" );
    exit( EXIT_SUCCESS );
  }
  
  comline.finalCheck( 0, "test [options]" );

  // defaults

  int itmax = 3;
  int msteps = DEF_MSTEPS;
  int rich = 0;
  real tol = DEF_TOL;
  int xprt_optim = 1;
  int gate_optim = 1;
  int debug = 0;
  int n_epochs = DEF_MAX_EPOCHS;
  int depth = 1;
  int n_in = n_input + 1;
  int n_experts = 1;
  printf("using %d experts\n", n_experts);

  real lambda = DEF_LAMBDA;
  real prior_var = DEF_PRIOR_VAR;
  
  if(strcmp(train_file, "-")==0)
    {
      error(0,"You must specify a training data file");
      comline.printUsage( "test [options]" );      
    }
  if(strcmp(out_file, "-")==0)
    {
      error(0,"You must specify an output weights file\n");
      comline.printUsage( "test [options]" );
    }
  char hostname[1024];
  time_t now;
  time(&now);
  assert(gethostname(hostname, 1024)==0);
  printf("Program %s version %s started on %s\n at %s \n",
		 progName, version,
		 hostname,ctime(&now));
  
  FILE* data_fp;

  data_fp = openFile(train_file, "r");

  Data* train_data_ptr;
  train_data_ptr = new Data(data_fp, npoints, n_input, n_output);
  Data& train_data = *train_data_ptr;

  real** orig_inputs = train_data.get_inputs(0);
  real** orig_targets = train_data.get_targets(0);

  // this section randomizes the inputs and targets


  int n_cv_train;
  int n_cv_test;

  real** rand_targ = RealMatrix2(npoints, n_output);
  real** rand_in = RealMatrix2(npoints, n_input+1);  
  int* map = IntVector(npoints);
  int fold = 0;
  time_t start_time;
  time_t stop_time;
  start_time = time(NULL);
  double time_per_pattern = 1.8765;
  double time_allowed = npoints * time_per_pattern;
  int BORED = 0;
  int n_folds = 50;
  real** trn_inp = RealMatrix2(npoints, n_input+1);
  real** trn_targ = RealMatrix2(npoints, n_output);
  real** tst_inp = RealMatrix2(npoints, n_input+1);
  real** tst_targ = RealMatrix2(npoints, n_output);
  if(seed == -1) seed = get_seed();
  set_seed(seed);

  while(((BORED == 0)||(fold < 3)) && (fold < n_folds)) {
	Reg_Tree* reg_tree;
	
	reg_tree = (Reg_Tree*) new Reg_Tree(npoints, 
										n_experts, 
										depth, 
										n_input, 
										n_output,
										debug,
										xprt_optim,
										gate_optim,
										tol,
										itmax,
										msteps,
										rich,
										seed,
										lambda,
										prior_var,
										0,0,1,1);
	
	randomize_matrix_rows2(orig_targets, 
						   npoints, 
						   n_output, map,
						   rand_targ);
	remap_matrix_rows2(orig_inputs, npoints, 
					   n_input+1, map, rand_in);

	real** inputs = rand_in;
	real** targets = rand_targ;

	n_cv_train = (int) (npoints * 2.0 / 3.0);
	n_cv_test = npoints - n_cv_train;
	printf("train = %d, test = %d\n", n_cv_train, n_cv_test);

	for(int n1 = 0; n1 < n_cv_train; n1 ++) {
	  for(int i = 0; i < n_input+1; i ++) {
		trn_inp[n1][i] = inputs[n1][i];
	  }
	  for(int o = 0; o < n_output; o ++) {
		trn_targ[n1][o] = targets[n1][o];
	  }
	}
	for(int n3 = n_cv_train; n3 < npoints; n3 ++) {
	  for(int i = 0; i < n_input+1; i ++) {
		tst_inp[n3 - n_cv_train][i] = inputs[n3][i];
	  }
	  for(int o = 0; o < n_output; o ++) {
		tst_targ[n3 - n_cv_train][o] = targets[n3][o];
	  }
	}
	int verb = 0;
	if(verb > 0) {
	printf("train\n");
	for(int n = 0; n < n_cv_train; n ++) {
	  printf("%d ", n);
	  for(int i = 0; i < n_input+1; i ++) {
		printf("%f ", trn_inp[n][i]);
	  }
	  for(int o = 0; o < n_output; o ++) {
		printf("%f ", trn_targ[n][o]);
	  }
	  printf("\n");
	}
	printf("\n");
	printf("test\n");
	for(int n2 = 0; n2 < n_cv_test; n2 ++) {
	  printf("%d ", n2);
	  for(int i = 0; i < n_input+1; i ++) {
		printf("%f ", tst_inp[n2][i]);
	  }
	  for(int o = 0; o < n_output; o ++) {
		printf("%f ", tst_targ[n2][o]);
	  }
	  printf("\n");
	}
	printf("\n");
	}		
	printf("Epoch\tNMSE\tDiff\n");
	int epochs = 0;

	char* extn = CharVector(1024);
	sprintf(extn, ".%d", fold);
	char* out_file_fold = CharVector(1024);
	strcpy(out_file_fold, out_file);
	strcat(out_file_fold, extn);
	FILE* out_fp = openFile(out_file_fold, "w");
	reg_tree->rand_posts(n_cv_train, trn_inp, trn_targ);
	epochs = reg_tree->cv_train(n_epochs+epochs, epochs, 
								trn_inp, trn_targ, 
								tst_inp, tst_targ,
								out_fp, n_cv_train,	n_cv_test);
	printf("\n saving params to file %s\n", out_file_fold);
	fclose(out_fp);
	FreeCharVector(&extn);
	FreeCharVector(&out_file_fold);
	delete reg_tree;
	stop_time = time(NULL);
	double time_elapsed = stop_time - start_time;
	if(time_elapsed > time_allowed)
	  BORED = 1;
	printf("time elapsed %f s\n time allowed %f s\n", 
		   time_elapsed, time_allowed);
	fold ++;
  }
  char* extn = CharVector(1024);
  strcpy(extn,".spec");
  char* spec_fn = CharVector(1024);
  strcpy(spec_fn, out_file);
  strcat(spec_fn, extn);
  FILE* spec_fp = openFile(spec_fn, "w");
  fprintf(spec_fp, "%d\n", fold);
  fclose(spec_fp);
  FreeRealMatrix2(&trn_inp);
  FreeRealMatrix2(&trn_targ);	
  FreeRealMatrix2(&tst_inp);
  FreeRealMatrix2(&tst_targ);

  time(&now);
  printf("Program %s ended at %s \n",progName,ctime(&now));
}


