#ifndef MACOPT_H
#define MACOPT_H

class Macopt 
{
public:

  Macopt(int _n, 
	 int _verbose = 0, 
	 int _rich = 0,
	 int _end_if_small_step = 0);
  Macopt() {};
  void macopt_init(int _n, int verb);
  void macopt_free();
  virtual ~Macopt() = 0;
  void macoptII(double *p,           // starting vector  
		int    n,            // number of dimensions
		int itmax,           // max iterations
		double tol           // tolerance   
		);
  void maccheckgrad(double *p,
		    int    n,
		    double epsilon, 
		    int stopat);	
protected:
  virtual double func(double* _p) = 0;
  virtual void dfunc(double* _p, double* _xi)  = 0;
//  double* params;
//  double* gradient;
  int a_n;     /* dimension of parameter space */
private:
  double maclinminII(double *p);
  double macprodII (double * , double * , double ) ;
  void macopt_restart ( int ) ;

private:
  int num_calls; // number of times optimiser is called.

   /* convergence declared when the gradient vector is smaller
		     in magnitude than this, or when the mean absolute 
		     step is less than this (see above) */
  double a_grad_tol_tiny ; /* if gradient is less than this, we definitely 
			    stop, even if we are not using a gradient 
			    tolerance */
  double a_step_tol_tiny ; /* if step is less than this, we stop, even if 
			    we are not using a step tolerance */
  int a_end_if_small_step ; /* defines the role of tol -- alternative is
			       end_on_small_grad */
  int a_rich ; /* whether to do the extra gradient evaluation at the beginning 
	      of each new line min */
  int a_verbose ;   /* if verbose = 1 then there is one report for each
		       line minimization.
		       if verbose = 2 then there is an additional 
		       report for
		       each step of the line minimization.
		       if verbose = 3 then extra debugging 
		       routines kick in.
		       */
  int a_linmin_maxits ; /* in maclinmin */
  double a_linmin_g1 ;  /* factors for growing and shrinking the interval */
  double a_linmin_g2 ;
  double a_linmin_g3 ;
  double a_lastx     ;      /* keeps track of typical step length */
  double a_lastx_default ;  /* if maclinmin is reset, lastx is set to this */
  
  /* These should not be touched by the user. 
     They are handy pointers for macopt
     to use 
     */
  double a_gtyp ; /* stores the rms gradient for linmin */
  double *a_pt , *a_gx , *a_gy , *a_gunused ;
  double *a_xi , *a_g , *a_h ; 

  int a_restart ;  /* whether to restart macopt - fresh cg directions */

  /* a_lastx :--- 1.0 might make general sense, (cf N.R.)
     but the best setting of all is to have 
     a prior idea of the eigenvalues. If 
     the objective function is equal to sum of N
     terms then set this to 1/N, for example 
     Err on the small side to be conservative. */
};

#endif

