#include "incs.h"
#include "net_class_ens.h"
#include "matrix.h"
#include "utils.h"
#include "functions.h"
#include "lu.h"

NET_Class_Ens::NET_Class_Ens(int n_batch, 
							 int num_ip, 
							 int num_op,
							 int dbg,
							 int _msteps,
							 float _tol,
							 int _hyper_epochs
							 )
  :    n_weights(num_ip * num_op),
       msteps(_msteps),
       tol(_tol),
       Net(n_batch, num_ip, num_op, dbg),
	   hyper_epochs(_hyper_epochs)
{
  weights = new double[n_weights];
  hessian = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    hessian[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      hessian[i][o] = new float[n_input];
    }
  }
  inv_hess = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    inv_hess[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      inv_hess[i][o] = new float[n_input];
    }
  }
  alpha = new double[n_output];
  for(int o = 0; o < n_output; o ++) {
    alpha[o] = 0.0;
  }
  ssw = new double[n_output];
  hess_trace = new float[n_output];
  uncertainty = new float[n_batch];
  strcpy(class_str, "NET_Class");
}

NET_Class_Ens::NET_Class_Ens(const NET_Class_Ens& old)
  :    n_weights(old.n_weights),
       msteps(old.msteps),
       tol(old.tol),
       Net(old),
	   hyper_epochs(old.hyper_epochs)
{
  weights = new double[n_weights];
  for(int i = 0; i < n_weights; i ++) {
	weights[i] = old.weights[i];
  }
  hessian = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    hessian[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      hessian[i][o] = new float[n_input];
    }
  }
  inv_hess = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    inv_hess[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      inv_hess[i][o] = new float[n_input];
    }
  }
  for(int o = 0; o < n_output; o ++) {
	for(int i = 0; i < n_input; i ++) {
	  for(int i2 = 0; i2 < n_input; i2 ++) {
		hessian[o][i][i2] = old.hessian[o][i][i2];
		inv_hess[o][i][i2] = old.inv_hess[o][i][i2];
	  }
	}
  }
  alpha = new double[n_output];
  for(int o = 0; o < n_output; o ++) {
    alpha[o] = old.alpha[o];
  }
  ssw = new double[n_output];
  for(int o = 0 ; o < n_output; o ++) {
	ssw[o] = old.ssw[o];
  }
  hess_trace = new float[n_output];
  for(int o = 0 ; o < n_output; o ++) {
	hess_trace[o] = old.hess_trace[o];
  }
  uncertainty = new float[size_batch];
  for(int n = 0; n < size_batch; n ++) {
	uncertainty[n] = old.uncertainty[n];
  }
  strcpy(class_str, "NET_Class");
}

NET_Class_Ens::NET_Class_Ens(int n_batch, FILE* fp) 
  : Net(n_batch, fp)
{
  n_weights = n_input * n_output;
  strcpy(class_str, "NET_Class");
  weights = new double[n_weights];

  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      float tmp;
      fscanf(fp, "%f ", &tmp);
      weights[i*n_input +j] = tmp;
    }
  }  
  hessian = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    hessian[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      hessian[i][o] = new float[n_input];
    }
  }
  inv_hess = new float**[n_output];
  for(int i = 0; i < n_output; i ++) {
    inv_hess[i] = new float*[n_input];
    for(int o = 0; o < n_input; o ++) {
      inv_hess[i][o] = new float[n_input];
    }
  }
  alpha = new double[n_output];
  for(int o = 0; o < n_output; o ++) {
    alpha[o] = 0.0;
  }
  hess_trace = new float[n_output];
  uncertainty = new float[n_batch];
}  
  
NET_Class_Ens::~NET_Class_Ens()
{
}

void NET_Class_Ens::randomize_weights(float offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
    weights[wt_num] = offset * 2.0* (get_drand() - 0.5);
  }
  /*  for(int o = 0; o < n_output; o ++) {
    alpha[o] = offset;
  }*/
}


void NET_Class_Ens::randomize_weights2(float offset)
{
  for(int wt_num = 0; wt_num < n_weights; wt_num ++) {
    weights[wt_num] = (1.0 + offset * 2.0*(get_drand() - 0.5)) * weights[wt_num]  ;

  }
}

void NET_Class_Ens::go_forward(float* in, float* out, int point)
{
  /*  if(n_output == 2) {
	  uncertainty[point] = 0.0;
	  for(int i = 0; i < n_input; i ++) {
      for(int i2 = 0; i2 < n_input; i2 ++) {
	  uncertainty[point] += in[i]*inv_hess[0][i][i2]*in[i2];
      }
	  }
	  out[0] = 0.0;
	  double arg = sqrt(1+ PI/8 * uncertainty[point]);
	  for(int ip = 0; ip < n_input; ip ++) {
      out[0] += weights[ip] * in[ip];
	  }
	  out[0] *= arg;
	  out[0] = 1.0/(1.0+Sexp(-out[0]));
	  out[1] = 1.0 - out[0];    
	  }*/
  //  else {
  for(int op = 0;  op < n_output;  op ++) {
    out[op] = 0.0;
  }
  for(int op = 0;  op < n_output;  op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      out[op] += weights[op*n_input + ip] * in[ip];
    }
  }
  log_softmax(out, out, n_output);
  for(int op = 0;  op < n_output;  op ++) {
	out[op] = Sexp(out[op]);
  }
 for(int op = 0;  op < n_output;  op ++) {
    outputs[point][op] = out[op];
  }
}
void NET_Class_Ens::go_forward_log(float* in, float* out, int point)
{
  /*  if(n_output == 2) {
	  uncertainty[point] = 0.0;
	  for(int i = 0; i < n_input; i ++) {
      for(int i2 = 0; i2 < n_input; i2 ++) {
	  uncertainty[point] += in[i]*inv_hess[0][i][i2]*in[i2];
      }
	  }
	  out[0] = 0.0;
	  double arg = sqrt(1+ PI/8 * uncertainty[point]);
	  for(int ip = 0; ip < n_input; ip ++) {
      out[0] += weights[ip] * in[ip];
	  }
	  out[0] *= arg;
	  out[0] = 1.0/(1.0+Sexp(-out[0]));
	  out[1] = 1.0 - out[0];    
	  }*/
  //  else {
  for(int op = 0;  op < n_output;  op ++) {
    out[op] = 0.0;
  }
  for(int op = 0;  op < n_output;  op ++) {
    for(int ip = 0; ip < n_input; ip ++) {
      out[op] += weights[op*n_input + ip] * in[ip];
    }
  }
  log_softmax(out, out, n_output);
  for(int op = 0;  op < n_output;  op ++) {
    outputs[point][op] = out[op];
  }
}
void NET_Class_Ens::go_forward(float* in, float* out, float* m2, int point)
{
  for(int op = 0;  op < n_output;  op ++) {
	m2[op] =  0.0;
  }
  go_forward(in, out, point);
  for(int op = 0;  op < n_output;  op ++) {
	 outputs[point][op] = out[op];
	 mom2[point][op] = m2[op];
   }
}

double NET_Class_Ens::train(int n_points, float** in, 
			float** targ)
{
  train_epochs ++;
  return train_params(n_points, in, targ);
}

double NET_Class_Ens::train(int n_points, 
			       float** in, 
			       float** targ,
			       float* wts)
{
  train_epochs ++;
  for(int n = 0; n < n_points; n ++) {
    jpost[n] = Sexp(wts[n]);
  }
  return train_params(n_points, in, targ, jpost);
}

void NET_Class_Ens::compute_gen_probs(int point, float* target)
{
  int index = max_index_vf(target, n_output);
  gen_prob[point] = Slog(outputs[point][index]);
}

void NET_Class_Ens::write_weights(FILE* fp)
{
  fprintf(fp, "%s\n", class_str);
  fprintf(fp, "%d %d\n", n_output, n_input);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      fprintf(fp, "%f ", weights[i*n_input +j]);
    }
    fprintf(fp, "\n");
  }
}

void NET_Class_Ens::read_weights(FILE* fp)
{
  int ninp;
  int noutp;
  fscanf(fp, "%s\n", class_str);
  fscanf(fp, "%d %d\n", &noutp, &ninp);
  for(int i = 0; i < n_output; i ++){
    for(int j = 0; j < n_input; j ++){
      float tmp;
      fscanf(fp, "%f ", &tmp);
      weights[i*n_input +j] = tmp;
    }
  }
}
void NET_Class_Ens::get_gradient(double* grad) {
   for(int o = 0; o < n_output; o ++) {
    for(int i = 0; i < n_input; i ++) {
      int n = o * n_input + i;
      grad[n] = alpha[o] * weights[n];
    }
  }
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], outputs[point], point);
    int wt_num = 0;
    for(int op = 0; op < n_output; op ++) {
      double err = (outputs[point][op] - targets[point][op]);
      for(int ip =0; ip < n_input; ip ++) {
		grad[wt_num] +=  jpost[point] * err * inputs[point][ip];
		wt_num ++;
      }
    }
  }
}
void NET_Class_Ens::get_gradient(double* grad, double* fn) {
  for(int o = 0; o < n_output; o ++) {
    for(int i = 0; i < n_input; i ++) {
      int n = o * n_input + i;
      grad[n] = alpha[o] * weights[n];
    }
  }
  *fn = 0.0;
  for(int o = 0; o < n_output; o ++) {
	ssw[o] = 0.0;
  }
  for(int o = 0; o < n_output; o ++) {
	for(int i = 0; i < n_input; i ++) {
	  int n = o * n_input + i;
	  ssw[o] += weights[n];
	}
  }
  for(int o = 0; o < n_output; o ++) {
	(*fn) += alpha[o]*ssw[o];
  }

  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], outputs[point], point);
    int wt_num = 0;
    for(int op = 0; op < n_output; op ++) {
      double err = (outputs[point][op] - targets[point][op]);
      for(int ip =0; ip < n_input; ip ++) {
		grad[wt_num] +=  jpost[point] * err * inputs[point][ip];
		wt_num ++;
      }
      (*fn) -= jpost[point] * targets[point][op] * Slog(outputs[point][op]);
    }
  }
}

double NET_Class_Ens::get_function() {
  double fn = 0.0;
 for(int o = 0; o < n_output; o ++) {
   ssw[o] = 0.0;
 }
 for(int o = 0; o < n_output; o ++) {
   for(int i = 0; i < n_input; i ++) {
	 int n = o * n_input + i;
	 ssw[o] += weights[n];
   }
 }
  for(int o = 0; o < n_output; o ++) {
	(fn) += alpha[o]*ssw[o];
  }
  for(int point = 0; point < batch_size; point ++) {
    go_forward(inputs[point], outputs[point], point);
    for(int op = 0; op < n_output; op ++) {
      fn -= jpost[point] * targets[point][op] * Slog(outputs[point][op]);
    }
  }
  return fn;
}
void NET_Class_Ens::compute_hessian(int n_points, 
				  float** in, 
				  float** out, 
				  float* wts,
				  float*** hess)
{
  for(int o = 0; o < n_output; o ++) {
    for(int i = 0; i < n_input; i ++) {
      for(int i2 = 0; i2 < n_input; i2 ++) {
	hess[o][i][i2] = 0.0;
      }
    }
    for(int i = 0; i < n_input; i ++) {
      hess[o][i][i] = alpha[o];
    }
  }
  for(int n = 0; n < n_points; n ++) {
    for(int o = 0; o < n_output; o ++) {
      float w = wts[n] * out[n][o] * (1.0 - out[n][o]);
      for(int i = 0; i < n_input; i ++) {
	for(int i2 = 0; i2 < n_input; i2 ++) {
	  hess[o][i][i2] += w  * in[n][i] * in[n][i2];
	}
      }
    }
  }
}
void NET_Class_Ens::update_hyperparameters(int n_points, 
										   float** in, 
										   float** targ,
										   float* wts) {
  inputs = in;
  targets = targ;
  go_forward_batch(n_points, in, outputs);
  for(int o = 0; o < n_points; o ++) {
	jpost[o] = Sexp(wts[o]);
  }
  compute_hessian(n_points, in, outputs, jpost, hessian);
  for(int o = 0; o < n_output; o ++) {
    hess_trace[o] = 0.0;
    for(int i = 0; i < n_input; i ++) {
      hess_trace[o] += inv_hess[o][i][i];
    }
  }
  for(int o = 0; o < n_output; o ++){
	double det = 0.0;
    det = lu_inverse(hessian[o], inv_hess[o], n_input);
  }
  for(int o = 0; o < n_output; o ++) {
     ssw[o] = 0.0;
	 for(int j = 0; j < n_input; j ++){
	   int n = o*n_input + j;
	   ssw[o] += weights[n] * weights[n];     
	 }
	 double alpha_0 = 0.1;
	 double ssw_0 = 0.1;
	 alpha[o] = (n_input + alpha_0) / (ssw[o] + ssw_0 + hess_trace[o]);
	 //	 printf("gate alpha = %f\n", alpha[o]);
  }
}

double NET_Class_Ens::get_joint(int point, float* targ)
{
  double joint = 0.0;

  for(int op = 0; op < n_output; op ++) {
	joint += targ[op] * Slog(outputs[point][op]);
	joint -= alpha[op] * ssw[op];
  }
  
  return joint;
}


void NET_Class_Ens::init(int n_points, float** in, 
						 float** targ, float* wts) {
  train(n_points, in, targ, wts);
}
