#ifndef NET_Regress_H
#define NET_Regress_H

#include "net.h"

class NET_Regress : public Net
{
public:
  NET_Regress(int n_batch, 
	      int num_ip, 
	      int num_op,
	      real _lambda,
	      real _prior_var,
	      int dbg = 0,
	      int _msteps = 10,
	      real _tol = 0.001);

  NET_Regress(const NET_Regress& old);
  NET_Regress(int n_batch, FILE* fp);
  ~NET_Regress();
  void go_forward(real* in, real* out);
  void go_forward_log(real* in, real* out);
  void go_forward(real* in, real* out, real* m2);
  void randomize_weights(real offset);
  void randomize_weights2(real offset);
  double train(int n_points, real** in, real** targ) ;
  double train(int n_points, real** in, real** targ, 
	       real* weights) ;
  void init(int n_points, real** in, real** targ, real* weights) ;
  void update_variance(int n_points, real** out, real** targ);
  void update_variance(int n_points, real** out, real** targ, real* wts);
  void update_hyperparameters(int n_points, real** in, 
							  real** targ,
							  real* weights);
  void write_weights(FILE* fp);
  void read_weights(FILE* fp);

  void get_gradient(double* grad);

  void get_gradient(double* grad, double* fn);
  double get_function();
  void compute_hessian(int n_points, 
		       real** in, 
		       real* wts,
		       real*** hess);
  double get_joint(real* in, real* targ);
  real get_prior(){ return 0;};
  double* get_variance() { return variance;};
  double* get_weights() { return weights;};
  void set_weights(real* wts) { weights = wts;};
  int get_num_weights() { return n_weights;};
  real get_evidence(int n_points, real** in, real** targ,
							real* wts);
  real get_sum_jposts(int n_points, real* wts);
protected:
  int n_weights;
  int msteps;
  real tol;

  double lambda;
  double prior_var;
  double* weights;
  double* last_weights;
  double* dweights;
  double* variance;
  double* inv_variance;
  double* var_acc;
};

#endif

