#include "incs.h"
#include "net_regress_conj.h"
#include "chol.h"
#include "matrix.h"
#include "utils.h"
#include "functions.h"
#include "conj.h"

NET_Regress_Conj::NET_Regress_Conj(int n_batch, 
				   int num_ip, 
				   int num_op,
				   int dbg,
				   int _itmax,
				   int _msteps,
				   int _rich,
				   float _tol,
				   float _lambda,
				   float _prior_var   )
  : itmax(_itmax),
    rich(_rich),
    NET_Regress(n_batch, num_ip, num_op, dbg, _msteps, _tol, _lambda,
		_prior_var),
    Conj(num_ip*num_op)
{
  ones_vec = new float[n_output * n_input];
  for(int i = 0; i < n_output*n_input; i ++) {
    ones_vec[i] = 1.0;
  }
  W = weights;
  strcpy(class_str,"NET_Regress_Conj");
}
NET_Regress_Conj::NET_Regress_Conj(const NET_Regress_Conj & old)
  : itmax(old.itmax),
    rich(old.rich),
    NET_Regress(old),
    Conj(old.n_input * old.n_output)
{
  ones_vec = new float[n_output * n_input];
  for(int i = 0; i < n_output*n_input; i ++) {
    ones_vec[i] = 1.0;
  }
  W = weights;
  strcpy(class_str,"NET_Regress_Conj");
}

NET_Regress_Conj::~NET_Regress_Conj()
{
}

double NET_Regress_Conj::train_params(int n_points, float** in, 
				      float** out, float** targ)
{
  return train(n_points, in, out, targ, ones_vec);
}

double NET_Regress_Conj::train_params(int n_points, 
				      float** in, 
				      float** out, 
				      float** targ,
				      float* wts)
{
  batch_size = n_points;
  inputs = in;
  outputs = out;
  targets = targ;
  float diff = 100;
  double cost = REALLY_VERY_BIG;
  int m_step = 0;
  int success = 0;
  int restart = 0;

  while((diff > tol) && (m_step < msteps)) {
    int iterations = itmax;
    int epochs = 0;
    if(m_step == 0) restart = 1;
    else restart = 0;
    double old_cost = cost;
    
    success = optimise(&iterations, &epochs, &restart, &cost);
    diff = old_cost  - cost;
    
    if(debug > 0)
      printf("conj:iter = %d/%d mstep = %d/%d ep = %d cost = %6.3f diff = %6.3g tol = %6.3f succ = %d restart = %d\n", 
	     m_step, msteps, iterations, itmax,  epochs, cost, diff, tol, success, restart);
    
    m_step ++;
     
  }

  return cost;
}

double NET_Regress_Conj::evaluate_gradient(double* dw) {
 for(int n = 0; n < NumWts; n ++) {
    dw[n] = 0.0;
  }
  double fn = 0.0;

  get_gradient(dw, &fn);

  return fn;
}
