#include "incs.h"
#include "utils.h"
#include "functions.h"
#include "macopt.h"
#include "net_regress_mac.h"

NET_Regress_Mac::NET_Regress_Mac(const NET_Regress_Mac & old)
  : itmax(old.itmax),
    rich(old.rich),
    NET_Regress(old),
    Macopt(old.n_input*old.n_output, old.debug, old.rich)
{
  old_weights = RealVector(n_output * n_input);
  mac_weights = RealVector(n_output * n_input + 1);
  grad = RealVector(n_output * n_input);
  for(int n1 = 0; n1 < n_output * n_input; n1 ++) {
	old_weights[n1] = old.old_weights[n1];
  }
  for(int n2 = 0; n2 < n_output * n_input+1; n2 ++) {
	mac_weights[n2] = old.mac_weights[n2];
  }
  for(int n = 0; n < n_output * n_input; n ++) {
	grad[n] = old.grad[n];
  }
  strcpy(class_str, "NET_Regress_Mac");
  if(debug > 0)
	printf("copying %s\n", class_str);
}

NET_Regress_Mac::NET_Regress_Mac(int n_batch, 
				 int num_ip, 
				 int num_op,
				 real _lambda,
				 real _prior_var,
				 int dbg,
				 int _itmax,
				 int _msteps,
				 int _rich,
				 real _tol)
  : itmax(_itmax),
    rich(_rich),
    NET_Regress(n_batch, num_ip, num_op, _lambda, _prior_var,
				dbg, _msteps, _tol),
    Macopt(num_ip*num_op, dbg, _rich)
{
  old_weights = RealVector(n_output * n_input);
  mac_weights = RealVector(n_output * n_input + 1);
  grad = RealVector(n_output * n_input);
  strcpy(class_str, "NET_Regress_Mac");
}

NET_Regress_Mac::NET_Regress_Mac(int n_batch, FILE* fp) 
: NET_Regress(n_batch, fp), 
  Macopt(n_input*n_output, 0, 0)
{
  old_weights = RealVector(n_output * n_input);
  mac_weights = RealVector(n_output * n_input + 1);
  grad = RealVector(n_output * n_input);
}  
  
NET_Regress_Mac::~NET_Regress_Mac()
{
  FreeRealVector(&old_weights);
  FreeRealVector(&mac_weights);
  FreeRealVector(&grad);
}


double NET_Regress_Mac::func(double* _weights) {
  double fn = 0.0;
  for(int w = 0; w < n_weights; w ++) {
	old_weights[w] = weights[w];
  }
  for(int j = 1; j <= a_n; j ++) {
    weights[j-1] = _weights[j];
  }
  fn = get_function();
  for(int w = 0; w < n_weights; w ++) {
	weights[w] = old_weights[w];
  }
  return fn;
}

double* NET_Regress_Mac::get_grad(int n_points, real** in,
								  real** targ) {
  batch_size = n_points;
  inputs = in;
  targets = targ;
  get_gradient(grad);
  return grad;
}
			   
void NET_Regress_Mac::dfunc(double* new_weights, double* dw) {
  for(int w = 0; w < n_weights; w ++) {
	old_weights[w] = weights[w];
  }

  for(int k = 1; k <= a_n; k ++) {
    weights[k-1] = new_weights[k];
  }

  get_gradient(grad);

  for(int j = 1; j <= a_n; j ++) {
    dw[j] = grad[j-1];
  }
  for(int w = 0; w < n_weights; w ++) {
	weights[w] = old_weights[w];
  }

}

double NET_Regress_Mac::train_params(int n_points)
{
  double cost = 0.0;
  batch_size = n_points;

  for(int n1 = 1; n1 <= a_n; n1 ++) {
    mac_weights[n1] = weights[n1-1];
  }
  // double epsilon=0.00001;
  //  printf("checking....\n");

  //  maccheckgrad( mac_weights, a_n, epsilon, 0 );

  macoptII(mac_weights, a_n, itmax, tol);

  for(int n = 1; n <= a_n; n ++) {
    weights[n-1] = mac_weights[n];
  }
  return cost;
}

double NET_Regress_Mac::train_params(int n_points, 
									 real* wts)
{
  batch_size = n_points;
  for(int p = 0; p < n_points; p ++)
	jpost[p] = wts[p];

  for(int n1 = 1; n1 <= a_n; n1 ++) {
    mac_weights[n1] = weights[n1-1];
  }
  double epsilon=0.00001;
  if(debug > 3) {
    printf("checking....\n");
    maccheckgrad( mac_weights, a_n, epsilon, 0 );
  }
  macoptII(mac_weights, a_n, itmax, tol);

  for(int n = 1; n <= a_n; n ++) {
    weights[n-1] = mac_weights[n];
  }
  double cost = 0.0;
  return cost;
}
