#include "incs.h"
#include "utils.h"
#include "net_utils.h"
#include "network.h"
#include "net.h"
#include "net_class_mac.h"
#include "net_regress_mac.h"
#include "net_class_ens_mac.h"
#include "net_regress_ens_mac.h"
#include "functions.h"


Network* get_new_net(Network* old_net) {
  Network* net = NULL;
  int good_net = 0;
  if(old_net == NULL) {
	error(1, "This net is NULL in get_new_net()");
  }
  char* net_class = old_net->get_class();
  if(strcmp(net_class, "NET_Class_Mac") == 0){
	net = new NET_Class_Mac(*((NET_Class_Mac*)old_net));
	good_net = 1;
  }
  if(strcmp(net_class, "NET_Regress_Mac") == 0) {
	NET_Regress_Mac* old_net2 = (NET_Regress_Mac*) old_net;
	net = new NET_Regress_Mac(*old_net2);
	good_net = 1;
  }
  if(strcmp(net_class, "NET_Class_Ens_Mac") == 0){
	net = new NET_Class_Ens_Mac(*((NET_Class_Ens_Mac*)old_net));
	good_net = 1;
  }
  if(strcmp(net_class, "NET_Regress_Ens_Mac") == 0) {
	NET_Regress_Ens_Mac* old_net2 = (NET_Regress_Ens_Mac*) old_net;
	net = new NET_Regress_Ens_Mac(*old_net2);
	good_net = 1;
  }
  if(good_net == 0)
	error(1, "No net created - bummer");
  return net;
}

Network* get_new_net(int size_batch, FILE* fp, char* net_class) {
  Network* net = NULL;
  int good_net = 0;
  if(strstr(net_class, "NET_Class_Ens") != NULL) {
	net = new NET_Class_Ens_Mac(size_batch, fp);
	good_net = 1;
  }
  else {
	if(strstr(net_class, "NET_Class") != NULL) {
	  net = new NET_Class_Mac(size_batch, fp);
	  good_net = 1;
	}
  }
  if(strstr(net_class, "NET_Regress_Ens") != NULL) {
	net = new NET_Regress_Ens_Mac(size_batch, fp);
	good_net = 1;
  }
  else {
	if(strstr(net_class, "NET_Regress") != NULL) {
	  net = new NET_Regress_Mac(size_batch, fp);
	  good_net = 1;
	}
  }
  if(good_net == 0)
	error(1, "No net created - bummer : %s", net_class);
  return net;
}

Network* get_new_class_net(int use_ensemble,
						   int optim_type,
						   int num_train,
						   int n_input,
						   int num_outputs,
						   int optim_iterations,
						   int msteps,
						   int optim_richness,
						   real optim_tolerance,
						   int hyper_epochs,
						   int debug) {
  Network* net = NULL;
  if(use_ensemble == 1) {
	net = new NET_Class_Ens_Mac(num_train, 
								n_input,
								num_outputs, 
								debug,
								optim_iterations,
								msteps,
								optim_richness,
								optim_tolerance
								);
  } 
  else 
	net = new NET_Class_Mac(num_train, 
							n_input, 
							num_outputs, 
							debug,
							optim_iterations,
							msteps,
							optim_richness,
							optim_tolerance
							);
  return net;
}


Network* new_class_net(int use_ensemble,
					   int num_train,
					   int n_input,
					   int num_outputs) {
  
  Network* net = NULL;
  if(use_ensemble) {
	net = new NET_Class_Ens_Mac(num_train, 
								n_input, 
								num_outputs);
  }
  else {
	net = new NET_Class_Mac(num_train, 
							n_input, 
							num_outputs);
  }
  return net;
}


Network* get_new_reg_net(int xprt_ensemble,
						 int xprt_optim,
						 int num_train,
						 int n_input,
						 int n_output,
						 int debug,
						 int optim_iterations,
						 int msteps,
						 int optim_richness,
						 real optim_tolerance,
						 real lambda,
						 real prior_var,
						 int xprt_hyper_epochs) {

  Network* expert = NULL;
  if(xprt_ensemble == 1) 
	expert = (NET_Regress_Ens_Mac*) 
	new NET_Regress_Ens_Mac(num_train,
							n_input,
							n_output,
							lambda,
							prior_var,
							debug,
							optim_iterations,
							msteps,
							optim_richness,
							optim_tolerance);
  else 
	expert = (NET_Regress_Mac*)
	new NET_Regress_Mac(num_train,
						n_input,
						n_output,
						lambda,
						prior_var,
						debug,
						optim_iterations,
						msteps,
						optim_richness,
						optim_tolerance
						);
  return expert;
}




void guess_class_scale(int n_points, int n_input, real** inputs, 
				 real** targets, real* scale) {

  NET_Class_Mac* net2 = new NET_Class_Mac(n_points,
										  n_input+1,
										  2);
  real** seg_targs = RealMatrix2(n_points, 2);
  real** seg_out = RealMatrix2(n_points, 2);
  for(int a1 = 0; a1 < n_points/2; a1++) {
	seg_targs[a1][0] = 0.1;
	seg_targs[a1][1] = 0.9;
  }	
  for(int a = n_points/2; a < n_points; a++) {
	seg_targs[a][1] = 0.1;
	seg_targs[a][0] = 0.9;
  }	

  net2->randomize_weights(1.0);
  double* w2 = net2->get_weights();
  int n_w2 = net2->get_num_weights();
  *scale = 0.0;
  for(int k2 = 0; k2 < n_w2; k2 ++) {
	*scale += w2[k2]*w2[k2];
  }
  for(int ep = 0; ep < 10; ep ++) {
	net2->train(n_points, inputs, seg_targs);
	net2->go_forward_batch(n_points, inputs, seg_out);
	double like = 0.0;
	for(int q = 0; q < n_points; q++) {
	  for(int op = 0; op < 2; op ++){
		like += seg_targs[q][op]*Slog(seg_out[q][op]);
	  }
	}
  }
  w2 = net2->get_weights();
  n_w2 = net2->get_num_weights();
  *scale = 0.0;
  for(int k3 = 0; k3 < n_w2; k3 ++) {
	*scale += w2[k3]*w2[k3];
  }
  *scale /= n_w2;
  *scale = sqrt(*scale);
 
}

int test_prune(real* joint_post, int n_points) {
  real sum = LZERO;
  for(int point = 0; point < n_points; point ++) {
	sum = LAdd(sum, joint_post[point]);
  }
  real thresh = Slog(MIN_PTS * n_points);
  if(sum < thresh) {
	return 1;
  }
  else return 0;
}
