#ifndef Tree_H
#define Tree_H

#include "incs.h"
#include <stddef.h>
#include "network.h"
#include "net.h"
#include "macopt.h"
#include "utils.h"
#include "functions.h"
#include "node_nt.h"
#include "node_term.h"
#include "net_utils.h"


class Tree : public Macopt
{
public:
  Tree(int n_batch, 
	   int num_expert, 
	   int num_depth,
	   int num_ip, 
	   int num_op,
	   int debug,
	   real _tol,
	   int _itmax,
	   int _msteps,
	   int _rich,
	   int _seed,
	   int _gate_ensemble = 0,
	   int _xprt_ensemble = 0,
	   int _gate_hyper_epochs = 1,
	   int _xprt_hyper_epochs = 1
	   );
  Tree(Tree *old);
  Tree(int n_batch, FILE* fp, int debug);
  Tree(const Tree& old);
  virtual ~Tree() = 0;
  int num_input();
  int num_output();

  void propagate_posts(Node** nodes, int num_nodes, int n);
  void randomize_weights(real xprt_offset, real gate_offset);
  void go_forward(real* in, real* out);
  void go_forward(real* in, real* out, real* m2);
  void go_forward_batch(int N, real** in, real** out);
  void go_forward_batch(int N, real** in, real** out, real** m2);  
  void update_all(int N, real** in, real** targ);
  void get_posts(int N, real** in, real** targ);
  void get_posts2(int N, real** in, real** targ);
  void get_ens_posts(int N, real** in, real** targ);
  real get_node_posts(Node* nd, int N, real** in, real** targ);
  real get_node_aux(Node* nd, int N, real** in, real** targ);
  real get_node_sum_jpost(Node* nd, int N, real** in, real** targ);
  void get_sa_posts(int N, real** in, real** targ);
  void get_hard_posts(int N, real** in, real** targ);
  void update_params(int N, real** in, real** targ);
  void update_hypers(int N, real** in, real** targ);
  int train(int n_epochs, int init_epoch, real** inputs, real** targets,
			FILE* fp_res, int ntrain);
  int train(int n_epochs, int init_epoch, real** inputs, real** targets,
			FILE* fp_res, int ntrain, int verb);
  int sa_train(int n_epochs, int init_epoch, real** inputs, real** targets,
			FILE* fp_res, int ntrain);
  double cv_train2(int n_epochs, 
			   int init_epoch,
			   real** trn_inputs, 
			   real** trn_targets,
			   real** tst_inputs, 
			   real** tst_targets,
			   FILE* fp_res, 
			   int ntrain,
			   int ntest,
				double min_err);
  int cv_train(int n_epochs, 
			   int init_epoch,
			   real** trn_inputs, 
			   real** trn_targets,
			   real** tst_inputs, 
			   real** tst_targets,
			   FILE* fp_res, 
			   int ntrain,
			   int ntest);
  int cv_train(int n_epochs, 
			   int init_epoch,
			   real** trn_inputs, 
			   real** trn_targets,
			   real** tst_inputs, 
			   real** tst_targets,
			   FILE* fp_res, 
			   int ntrain,
			   int ntest,
			   real* min_err);
  int cv_sa_train(int n_epochs, 
			   int init_epoch,
			   real** trn_inputs, 
			   real** trn_targets,
			   real** tst_inputs, 
			   real** tst_targets,
			   FILE* fp_res, 
			   int ntrain,
			   int ntest);
  int ens_train(int n_epochs, 
				int init_epoch,
				int hyp_epochs,
				real** inputs, 
				real** targets,
				FILE* fp_res, 
				int ntrain);
  int train_params(int n_epochs, 
				   int init_epoch,
				   real** inputs, 
				   real** targets,
				   FILE* fp_res, 
				   int ntrain);

  void rand_nt_node(Node_NT* nd, int N);
  void rand_params_nt_node(Node_NT* nd);
  void reorder_nt_node(Node_NT* nd, int N);
  real get_gen_prob(int);
  void print_posteriors(int N);
  void write_weights(char* file_name);
  void read_weights(char* fn);
  void write_weights(FILE* fp);
  void read_weights(FILE* fp);
  void count_nodes(Node* nd, int* n_terms, int* n_nonterms);
  void collect_nodes(Node* nd);
  void delete_nodes();
  void count_params(Node* nd, int* n_params);
  int get_n_params();
  void gather_nodes();
  virtual double compute_like(int N, real** out, real** targ) = 0;
  virtual double compute_error(int N, real** out, real** targ) = 0;
  virtual Node* build_recursive_tree(Node* parent,
									 int num_experts, 
									 int* current_depth, 
									 int num_train, 
									 int xprt_optim,
									 int gate_optim) = 0;
  int get_new_id();

  Node* build_recursive_tree(Node_NT* old, Node* parent);
  Node* build_recursive_tree(Node_NT& old, Node* parent);

  void rand_posts(int N, real** in, real** targ);
  void ord_init(int N, real** in, real** targ);
  void wta_train(int N, real** in, real** targ, int* tc);
  void wta_init(int N, real** in, real** targ);

  void rand_terms(int N, real** in, real** targ);
  void rand_gates();
  void write_predictions(FILE* fp, int N, real** pred);

  // functions for global training tree_mac.cc

  void dfunc(double* new_weights, double* dw);
  double func(double* _weights);
  double mac_train(int n_points, real** in, real** targ);
  double get_function();
  void get_gradient(double* g);

  // functions for merging tree_prune.cc
  void prune_tree(int n_points, real** in, real** targ, FILE* fp_out);
  int merge_nodes(int n_points, real** in, real** targ);
  void prune_tree(int n_train, real** trn_in, real** trn_targ,
				  int n_test, real** tst_in, real** tst_targ, FILE* fp_out);
  int merge_nodes(int n_train, real** trn_in, real** trn_targ,
				  int n_test, real** tst_in, real** tst_targ);
  
  // functions for growing tree_grow.cc
  
  
  Node* create_new_nt(Node* old_child, int num_children, Network* old_gate);
  void delete_new_nt(Node_NT* nd);
  void split_nodes_hard(int n_points, real** in, real** targ);
  void split_nodes_fsearch(int n_points, real** in, real** targ);
  void split_nodes_randomly(int n_points, real** in, real** targ);
  void split_nodes_max_like(int n_points, real** in, real** targ);
  Node_NT* split_this_node(real best_thresh, int best_feat, 
							Node* old_nd, int n_points, 
							real** jpost, real** cpost, 
							real** in, real** targ);
  void grow_train(int n_epochs, 
				  real** inputs,
				  real** targets,
				  FILE* fp_res,
				  int ntrain,
				  int n_splits);
  void grow_train(int n_epochs, 
				  int n_train, real** trn_inputs, real** trn_targets, 
				  int n_test, real** tst_inputs, real** tst_targets, 
				  FILE* fp_out);

  void swap_nodes(Node* new_node, Node* old_node);
  void assign_probs_from_thresh(Node* nd, int n_points,
							  real** in, real best_thresh,
							  int best_feat,
							  real** jpost, real** cpost
							  );
  void assign_probs_from_thresh_sig(Node* nd, int n_points,
									real** in, real best_thresh,
									int thresh_posn,
									int best_feat,
									real** jpost, real** cpost,
									int n_batch,
									real** srt_in,
									int* indx
									);
  real find_best_hard_thresh(Node* nd,
							 int batch_n, real** in_batch, 
							 real** in, real** targ,
							 int* best_feat,
							 real* max_diff,
							 int* best_thresh_posn);

  Node_NT* find_best_soft_node(Node* nd, int n_points, 
							   real** in, real** targ, 
							   real* glob_best_diff,
							   int* success);

  Node_NT* find_best_hard_node(Node* nd, int n_points,
							   real** in, real** targ,
							   real* diff,
							   int* success);

  Node_NT* pick_node_rand(Node* nd, int n_points,
							  real** jpost, real** cpost,
							  real** in, real** targ,
							  real* diff);
  void grow_ens_train(int n_epochs, real** inputs, real** targets, 
					  FILE* fp_res, int npoints);
  real get_evidence(int n_points, real** in, real** targ);
protected:
  Node* node;
  int size_batch;
  int n_conn;
  int n_expert;
  int n_input;
  int n_output;
  int n_levels;
  int debug;
  real optim_tolerance;
  int optim_iterations;
  int msteps;
  int optim_richness;
  double like;
  double aux;
  char* class_str;

  real node_xp_offset;
  real node_gt_offset;
  
  int seed;
  int term_count;
  int n_terminals;
  int gate_count;
  int n_gates;
  int n_nodes;

  int gate_ensemble;
  int xprt_ensemble;
  int gate_hyper_epochs;
  int xprt_hyper_epochs;
  real xprt_rand_offset;
  real gate_rand_offset;
  real beta;
  real** outputs;
  real** tmp_ops;
  real*  tmp_op;  
  real** tmp_mom;
  Network** terminals;
  Node** term_nodes;
  Network** gates;
  Node** nt_nodes;
  real** mac_inputs;
  real** mac_targets;
  real* mac_weights;
  real* old_weights;
  real* weights;
  real* grad;
  int n_weights;
  int batch_size;
  int itmax;
  real tol;
  int mac_pass;
  void get_weights(real* weights);
  void set_weights(real* weights);
  void rescan_tree();
  real** srt_in;
  real** jpost;
  real** cpost;
  real* tmp;
  int* indx;
  real** in_batch;
};

#endif
