#include "tree.h"

void Tree::get_ens_posts(int n_points, real** in, real** targ)
{

  real over_sum = LZERO;
  real lik = 0.0;
  real jp_sum = LZERO;
  real ax = 0.0;
  for(int n = 0; n < n_points; n ++) {
	node->set_joint_prob(0.0, in[n], n);
	real sum = LZERO;
	real lik_n = LZERO;
	for(int ch = 0; ch < n_terminals; ch ++) {
	  real dens = terminals[ch]->get_joint(in[n], targ[n]);
	  real j = term_nodes[ch]->get_joint_prob(n);
	  jp_sum = LAdd(jp_sum, j);
	  real jp = j + dens;

	  lik_n = LAdd(lik_n, jp);
	  term_nodes[ch]->set_joint_post(n, jp);
	  if(jp > LSMALL)
		sum = LAdd(sum, jp);
	}
	lik += lik_n;
	for(int ch2 = 0; ch2 < n_terminals; ch2 ++) {	
	  real jp = term_nodes[ch2]->get_joint_post(n) - sum;
	  term_nodes[ch2]->set_joint_post(n, jp);
	  real dens = terminals[ch2]->get_joint(in[n], targ[n]);
	  real j = term_nodes[ch2]->get_joint_prob(n);
	  ax += Sexp(jp) * (j + dens);
	  over_sum = LAdd(over_sum, jp);
	}
	propagate_posts(term_nodes, n_terminals, n);
  }
  for(int g = 0; g < n_gates; g++) {
	real p = gates[g]->get_prior();
	ax += p;
	lik += p;
  }
  for(int e = 0; e < n_terminals; e ++) {
	real p = terminals[e]->get_prior();
	ax += p;
	lik += p;
  }
  aux = ax;
  like = lik;
}


int Tree::ens_train(int max_epochs,
					int init_epoch,
					int hyp_epochs,
					real** inputs, 
					real** targets,
					FILE* fp_out, 
					int ntrain) {
  time_t start_time;
  time_t stop_time;
  start_time = time(NULL);
  double time_per_pattern = 1.8765;
  double time_allowed = ntrain * time_per_pattern / 4.0;

  double log_like = 0.0;
  double aux_like = 0.0;
  real diff = 0.0;
  int epoch = 0;
  fprintf(stdout, "Ens train started\n");
  fflush(stdout);

  int epochs_allowed = hyp_epochs;

  train_params(epochs_allowed, 0, inputs, targets, fp_out, ntrain);
  get_ens_posts(ntrain, inputs, targets);
  log_like = like;
  aux_like = aux;
  double err = compute_error(ntrain, inputs, targets); 
  double ddiff;
  real old_diff = 0.0;
  real average_diff = 2.0;
  int average_length = 3;

  fprintf(stdout,
		  "%4d\t%4.3g\t%4.3g\t%4.3g\t%4.3g\n",
		  0,
		  err,
		  log_like,
		  aux_like,
		  0.0);
  fflush(stdout);
  real av = 0.0;
  int BORED = 0;

  while((((epoch < 6) || (BORED == 0))&&(epoch < max_epochs)&&(average_diff) > 1e-4)) {
    if(epoch > 1) {
      if((epoch % average_length) == 0){
		average_diff = av / average_length;
		old_diff = average_diff;
		av = 0.0;
      }
      av += diff;
    }
	update_hypers(ntrain, inputs, targets);
	epoch ++;
	get_ens_posts(ntrain, inputs, targets);
    double old_log_like = like;
	double old_aux_like = aux;
	int ep = train_params(hyp_epochs, 0, inputs, 
						  targets, fp_out, ntrain);
	get_ens_posts(ntrain, inputs, targets);
	log_like = like;
	aux_like = aux;
	double err = compute_error(ntrain, inputs, targets);
	old_diff = diff;
    diff = fabs((log_like - old_log_like) / old_log_like);
	ddiff = old_diff - diff;
	fprintf(stdout, 
			"%4d\t%4.3g\t%4.3g\t%4.3g\t%4.3g\t%4.3g\n",
		   epoch,
		   err,
		   log_like,
		   aux_like,
		   diff,
		   average_diff);
	fflush(stdout);
	stop_time = time(NULL);
	double time_elapsed = stop_time - start_time;
	printf("time allowed %f s, time elapsed %f s\n", time_allowed, 
		   time_elapsed);
	if(time_elapsed > time_allowed) {
	  BORED = 1;
	  printf("I got bored\n");
	}
  }
  return epoch;
}



void Tree::update_hypers(int n_points, real** in, real** targ)
{
  get_posts(n_points, in, targ);
  if(debug > 3){
    for(int point2 = 0; point2 < n_points; point2 ++){
      node->print_posteriors(point2);
    }
  }
  node->update_hypers(n_points, in, targ);
}


int Tree::train_params(int n_epochs, 
					   int init_epoch,
					   real** inputs, 
					   real** targets,
					   FILE* fp_out, 
					   int ntrain) {
  double log_like = 0.0;
  double aux_like = 0.0;
  real old_diff = 0.0;
  real average_diff = 2.0;
  int average_length = 3;
  double diff = 0.0;
  real av = 0.0;
  int rec_epoch = init_epoch;
  int epoch = 0;
  while(((fabs(average_diff) > 1e-5) && (epoch < n_epochs))) {
    if(epoch > 1) {
      if((epoch % average_length) == 0){
		average_diff = av / average_length;
		old_diff = average_diff;
		av = 0.0;
      }
      av += diff;
    }
	get_posts(ntrain, inputs, targets);
    double old_log_like = like;
    update_params(ntrain, inputs, targets);
	get_posts(ntrain, inputs, targets);
	log_like = like;
	aux_like = aux;
	double error = compute_error(ntrain, inputs, targets);
    diff = fabs(log_like - old_log_like);
	if(diff < 0.01) {
	  get_posts(ntrain, inputs, targets);
	  mac_train(ntrain, inputs, targets);
	}
	get_ens_posts(ntrain, inputs, targets);
	double post = like;
	int verb = 1;
	if(verb > 0) {
	  fprintf(stdout,
			  "%4d/%4d\t%4.3g\t%4.3g\t%4.3g\t%4.3g\t%4.3g\n",
			 rec_epoch,
			 n_epochs,
			 error,
			 log_like,
			 aux_like,
			  post,
			 diff);
	  fflush(stdout);
	}
    epoch ++;
	rec_epoch ++;
  }
  return rec_epoch;
}

void Tree::update_params(int n_points, real** in, real** targ)
{
  get_posts(n_points, in, targ);
  if(debug > 3){
    for(int point2 = 0; point2 < n_points; point2 ++){
      node->print_posteriors(point2);
    }
  }
  node->update_params(n_points, in, targ);
}


real Tree::get_evidence(int n_points, real** in, real** targ) {
  real evidence = 0.0;
  for(int t = 0; t < n_terminals; t ++) {
	real* jp = term_nodes[t]->get_jpost_vec();
	evidence += terminals[t]->get_evidence(n_points, in, targ, jp);
  }
  for(int nt = 0; nt < n_gates; nt ++) {
	real* jp = nt_nodes[nt]->get_jpost_vec();
	Node_NT* nd = (Node_NT*)nt_nodes[nt];
	nd->get_new_gate_targets(n_points);
	real** gt = nd->get_gate_targets();
	evidence += gates[nt]->get_evidence(n_points, in, gt, jp);
  }
  return evidence;
}
