#include "tree.h"

void Tree::rescan_tree() {
// now fix all old gates etc. calling gather.
  delete [] term_nodes;
  delete [] terminals;
  delete [] gates;
  delete [] nt_nodes;
  gather_nodes();
  mac_pass = 0;
  macopt_free();
  int np = get_n_params();
  //  printf("nparams = %d\n", np);
}

void Tree::prune_tree(int n_train, real** trn_in, real** trn_targ,
					  int n_test, real** tst_in, real** tst_targ,
					  FILE* fp_out) {
  int good_merge = 1;
  double base_err = compute_error(n_test, tst_in, tst_targ);
  rewind(fp_out);
  write_weights(fp_out);
  while(good_merge) {
	good_merge = merge_nodes(n_train, trn_in, trn_targ,
							 n_test, tst_in, tst_targ);
	double err = compute_error(n_test, tst_in, tst_targ);	
	if(err < base_err) {
	  printf("err %f < best_err %f : saving weights \n", err, base_err);
	  base_err = err;
	  rewind(fp_out);
	  write_weights(fp_out);
	}
  }
}

int Tree::merge_nodes(int n_train, real** trn_in, real** trn_targ,
					  int n_test, real** tst_in, real** tst_targ) {
  printf("merging...\n");
  double base_err = compute_error(n_test, tst_in, tst_targ);
  double min_err = FLT_MAX;
  int best_merge = -1;
  Node* best_node = NULL;
  Network* default_expert = ((Node_Term*)term_nodes[0])->get_expert();
  get_posts2(n_train, trn_in, trn_targ);
  int max_nodes = n_gates;
  Node_NT** node_array = new Node_NT*[max_nodes];
  for(int nt = 0; nt < max_nodes; nt ++) {
	node_array[nt] = (Node_NT*) nt_nodes[nt];
  }
	
  for(int nt = 0; nt < max_nodes; nt ++) {
	Node_NT* nd = (Node_NT*) node_array[nt];
	  Node* old_nd = (Node*)node_array[nt];
	  Node_Term* new_term = new Node_Term(*old_nd, default_expert);
	  new_term->init(n_train, trn_in, trn_targ);

	  fflush(stderr);
	  // now swap it 
	  //	  printf("swapping nodes... n_gates = %d\n", n_gates);
	  if(old_nd->get_id() != 1) {
		swap_nodes((Node*) new_term, old_nd);
		rescan_tree();

		double tst_err = compute_error(n_test, tst_in, tst_targ);
		
		if(tst_err < min_err) {
		  min_err = tst_err;
		  best_merge = nt;
		  best_node = new_term;
		}
		//		printf("node %d %d err = %f best_err = %f\n", old_nd->get_id(), nt, tst_err, min_err);
		// swap back
		swap_nodes(old_nd, (Node*) new_term);
		rescan_tree();
	  }
	}
  //  }
  if(best_merge != -1) {
	printf("best node = %d, err = %f last_err = %f\n", best_merge, min_err,
		   base_err);
	if(n_gates == 1) 
	  node = best_node;
	else
	  swap_nodes(best_node, (Node*)node_array[best_merge]);
	printf("merged node %d\n", best_node->get_id());
	rescan_tree();
	
	return 1;
  }
  else {
	printf("no good merges\n");
	return 0;
  }
}

